#include <clipboard>
#include <funcs>
#include <memory>
#include <sqlite>

class Clip {
formats := []

;adds a format from the clipboard to our list of formats.
add_format_from_clipboard(format, data, size) {
this.formats.insert(new ClipEntry(data, size, format))
}

find_format(f) {
desc := ""
for i, format in this.formats {
if (format.format == f) {
desc := format
break
}
}
return desc
}

;frees everything
free() {
for i, format in this.formats {
format.free()
}
loop, % this.formats.MaxIndex()
{
this.formats.remove()
}
}

to_clipboard(cb) {
for i, format in this.formats {
format.to_clipboard(cb)
}
}

getSize() {
size := 0
for i, format in this.formats
size := size + format.size
return size
}

  update_time() {
    db.exec("update items set time=" . time() . " where id=" . id)
  }

}

class ClipEntry {
__new(data_ptr, size, format) {
this.data_ptr := data_ptr
this.size := size
this.format := format
}

free() {
memory_free(this.data_ptr)
this.data_ptr := 0
}

to_clipboard(cb) {
cb.set_clipboard_data(this.format, this.data_ptr, this.size)
}

}

class ClipDb {
__new(file) {
global schema
cs=0
if (fileExist(file) ) {
fileGetSize size, %file%
if (size == 0)
cs=1
} else {
cs=1
}

this.db := open_db(file)
sqlite_exec(this.db, "pragma foreign_keys=on")
if (cs) {
this.exec("create table items (id integer primary key, size integer, description text, time integer)")
this.exec("create index idx_time on items(time)")
this.exec("create index idx_size_desc on items (size, description)")
this.exec("create table formats ( clip_id integer, format text, data blob, foreign key (clip_id) references items (id) on delete cascade)")
this.exec("create index idx_clip_id on formats(clip_id)")
}
this.additem := prepare(this.db, "insert into items values(null, ?, ?, ?)")
this.addformat := prepare(this.db, "insert into formats values(?, ?, ?)")
this.query := prepare(this.db, "select id, substr(description, 1, 259) from items order by time")
this.formats_for_id := prepare(this.db, "select format, data from formats where clip_id=?")
this.dupcheck := prepare(this.db, "select id from items where size=? and description=?")

this.exec("pragma synchronous=off")

}

exec(sql) {
return sqlite_exec(this.db, sql)
}

store(clip) {
global CF_UNICODE
;store a clip
size := clip.getSize()
time := time()
;begin transaction, store the main item
this.exec("begin")
sqlite3_clear_bindings(this.additem)
desc := clip.find_format(CF_UNICODE)
if (!isobject(desc))
msgbox no desc
sqlite3_bind_int(this.additem, 1, size)
sqlite3_bind_text16(this.additem, 2, desc.data_ptr, -1)
sqlite3_bind_int(this.additem, 3, time)
sqlite3_exec_prepared(this.db, this.additem)
sqlite3_reset(this.additem)
rowid := sqlite3_last_insert_rowid(this.db)
;Now add all the formats
for i, format in clip.formats
{
sqlite3_bind_int(this.addformat, 1, rowid)
name := clipboard_format_to_name(format.format)
sqlite3_bind_text16(this.addformat, 2, &name, -1)
sqlite3_bind_blob(this.addformat, 3, format.data_ptr, format.size)
sqlite3_exec_prepared(this.db, this.addformat)
sqlite3_reset(this.addformat)
}
this.exec("commit")
}

getListItems() {
lst := []
while (sqlite3_step(this.query) == 100) {
id := sqlite3_column_int64(this.query, 0)
text := sqlite3_column_text16(this.query, 1)
lst.insert([id,text])
}
sqlite3_reset(this.query)
return lst
}

remove_clip(id) {
this.exec("delete from items where id=" . id)
}

;load clip id from db
clip_from_db(id) {
c := new clip()
sqlite3_bind_int64(this.formats_for_id, 1, id)
while (sqlite3_step(this.formats_for_id) == 100) {
format_name := sqlite3_column_text16(this.formats_for_id, 0)
size := sqlite3_column_bytes(this.formats_for_id, 1)
blob := sqlite3_column_blob(this.formats_for_id, 1)
data := memory_allocate(size)
dllCall("RtlMoveMemory", "ptr", data, "ptr", blob, "uint", size)
format := create_clipboard_format_if_needed(format_name)
e := new ClipEntry(data, size, format)
c.formats.insert(e)
}
sqlite3_reset(this.formats_for_id)
return c
}

;check if this clip already exists in our db, mostly exact match
exists(c) {
global CF_UNICODE
result := 0
desc := c.find_format(CF_UNICODE)
sqlite3_bind_int64(this.dupcheck, 1, c.getSize())
sqlite3_bind_text16(this.dupcheck, 2, desc.data_ptr, -1)
while (sqlite3_step(this.dupcheck) == 100) {
result := sqlite3_column_int64(this.dupcheck, 0) ;clips with the same size and description are duplicates
break
;todo: read into memory and compare each format if we need to
}
sqlite3_Reset(this.dupcheck)

return result
}
}
